import { Float, Mat3, Mat4, ReadonlyMat3, ReadonlyMat4, Vec4 } from '@andykswong/munum-types';
/**
 * Create a new identity {@link Mat4}.
 * @returns identity mat4
 */
export declare function create(): Mat4;
/**
 * Set a {@link Mat4} to identity.
 * @returns out = I4
 */
export declare function id(out: Mat4): Mat4;
/**
 * Create a {@link Mat4} from {@link ReadonlyMat3}.
 * @param m input mat3
 * @param out output mat4
 * @returns out = [M 0 0 1]
 */
export declare function fromMat3(m: ReadonlyMat3, out?: Mat4): Mat4;
/**
 * Create a {@link Mat3} from upper-left 3x3 of a {@link ReadonlyMat4}.
 * @param m input mat4
 * @param out output mat3
 * @returns out = mat3(M)
 */
export declare function toMat3(m: ReadonlyMat4, out?: Mat3): Mat3;
/**
 * Copy a {@link Mat4}.
 * @returns out
 */
export declare const copy: (v: readonly [m11: number, m21: number, m31: number, m41: number, m12: number, m22: number, m32: number, m42: number, m13: number, m23: number, m33: number, m43: number, m14: number, m24: number, m34: number, m44: number], out?: Mat4) => Mat4;
/**
 * Sum 2 {@link Mat4}.
 * @returns out = a + b
 */
export declare const add: (a: readonly [m11: number, m21: number, m31: number, m41: number, m12: number, m22: number, m32: number, m42: number, m13: number, m23: number, m33: number, m43: number, m14: number, m24: number, m34: number, m44: number], b: readonly [m11: number, m21: number, m31: number, m41: number, m12: number, m22: number, m32: number, m42: number, m13: number, m23: number, m33: number, m43: number, m14: number, m24: number, m34: number, m44: number], out?: Mat4) => Mat4;
/**
 * Subtract 2 {@link Mat4}.
 * @returns out = a - b
 */
export declare const sub: (a: readonly [m11: number, m21: number, m31: number, m41: number, m12: number, m22: number, m32: number, m42: number, m13: number, m23: number, m33: number, m43: number, m14: number, m24: number, m34: number, m44: number], b: readonly [m11: number, m21: number, m31: number, m41: number, m12: number, m22: number, m32: number, m42: number, m13: number, m23: number, m33: number, m43: number, m14: number, m24: number, m34: number, m44: number], out?: Mat4) => Mat4;
/**
 * Multiply a {@link Mat4} by a constant.
 * @returns out = s * M
 */
export declare const scale: (m: readonly [m11: number, m21: number, m31: number, m41: number, m12: number, m22: number, m32: number, m42: number, m13: number, m23: number, m33: number, m43: number, m14: number, m24: number, m34: number, m44: number], s: Float, out?: Mat4) => Mat4;
/**
 * Transpose a {@link Mat4}.
 * @returns [M]T
 */
export declare const tr: (m: readonly [m11: number, m21: number, m31: number, m41: number, m12: number, m22: number, m32: number, m42: number, m13: number, m23: number, m33: number, m43: number, m14: number, m24: number, m34: number, m44: number], out?: Mat4) => Mat4;
/**
 * Multiply 2 {@link Mat4}.
 * @returns out = a * b
 */
export declare const mul: (a: readonly [m11: number, m21: number, m31: number, m41: number, m12: number, m22: number, m32: number, m42: number, m13: number, m23: number, m33: number, m43: number, m14: number, m24: number, m34: number, m44: number], b: readonly [m11: number, m21: number, m31: number, m41: number, m12: number, m22: number, m32: number, m42: number, m13: number, m23: number, m33: number, m43: number, m14: number, m24: number, m34: number, m44: number], out?: Mat4) => Mat4;
/**
 * Multiply a {@link Mat4} with a {@link Vec4}.
 * @returns out = m * v
 */
export declare const vmul: (a: readonly [m11: number, m21: number, m31: number, m41: number, m12: number, m22: number, m32: number, m42: number, m13: number, m23: number, m33: number, m43: number, m14: number, m24: number, m34: number, m44: number], b: readonly [x: number, y: number, z: number, w: number], out?: Vec4) => Vec4;
/**
 * Returns inverse of a {@link Mat4}, or null if matrix is not invertible.
 * @param a input matrix
 * @param out output matrix
 * @returns out = M^-1, or null if matrix is not invertible
 */
export declare function invert(a: ReadonlyMat4, out?: Mat4): Mat4 | null;
/**
 * Returns the 3x3 normal matrix, which is the inverse transpose matrix, from a {@link ReadonlyMat4}.
 * Returns null if original matrix is not invertible.
 * @param a input matrix
 * @param out output matrix
 * @returns out = (mat3(M)^-1)T, or null if matrix is not invertible
 */
export declare function normMat3(a: ReadonlyMat4, out?: Mat3): Mat3 | null;
//# sourceMappingURL=mat4.d.ts.map